#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>    /* for FIONBIO */
#include <netinet/in.h>   /* struct in_addr (internet address), struct
                             sockaddr_in,  struct ip_mreq */
#include <arpa/inet.h>    /* for inet_* calls */
#include <netdb.h>        /* for gethostbyaddr call */
#include "cdefs.h"
#include "sourcenic.h"

/* setsockopt and getsockopt return 0 on success and -1 on error */

/* mcaddress is the multicast to join and
   ifaddress is the IP address of the local interface to look for the
             multicast data.
   Both are char string of form (for example) 192.168.1.200 */
int opensocket (int *port, int prange, char *mcaddress, char *ifaddress,
                int sbufsize, int debug) {
  int sock;
  int status;
  int unblock=0;  /* Set to 1 to make requests unblocking, 0 for blocking */
                  /* Set to 0 for blocking so CPU conserved */
  socklen_t len;
  int nport    = *port;
  int lastport = *port + prange;
  unsigned int retbufsize;
  unsigned int bufsize=sbufsize;  /* Try to get this much -> kernel will tell
                                       us how much we can actually get! */
  
  unsigned char multicast_loop = 0;  /* Disable multicast looping */
  struct ip_mreq mreq;               /* Multicast request structure */
  struct sockaddr_in bindsockaddress;

  in_addr_t if_addr;
  socklen_t if_siz = sizeof(if_addr);

  /*
  struct ip_mreq 
  {
          struct in_addr imr_multiaddr;    IP multicast address of group
          struct in_addr imr_interface;    local IP address of interface
  };  */


  /* Convert network addresses to binary form and place in mreq structure */
  inet_aton(mcaddress,&mreq.imr_multiaddr);
  inet_aton(ifaddress,&mreq.imr_interface);

  /* Below command opens datagram socket, AF_INET means
     IPv4 Internet protocols (see man ip).
     The other valid value for 3rd argument is 0 (shouldn't matter tho) */
  if (debug > 0) printf("About to open UDP socket\n");   
  sock = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (sock<0) { perror("Err opening socket\n"); return -1; }
  if (debug > 0) printf("Opened socket %d\n",sock);

  TRY_NEXT_PORT:
  bindsockaddress.sin_family      = AF_INET;
  bindsockaddress.sin_port        = htons(nport);
  /* bindsockaddress.sin_addr.s_addr = INADDR_ANY; */
  bindsockaddress.sin_addr.s_addr = mreq.imr_multiaddr.s_addr;
  *port = nport;

  /* Disable multicast looping -> by default it is enabled 
     When enabled it will loop back and send data to host (so more than
     one user can listen).  Right now we only want one user listening (us!)
     so we will disable */
  status = setsockopt(sock,IPPROTO_IP,IP_MULTICAST_LOOP,&multicast_loop,
                      sizeof(multicast_loop));
  if (status<0) { perror("Err Disabling multicast looping\n"); return -6; }
  
  /* Now become a member of multicast group! */
  status = setsockopt(sock,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq));
  if (status<0) { perror("Err joining multicast group\n"); return -7; }

  /* Bind socket */
  status = bind (sock, (struct sockaddr *)&bindsockaddress,
                 sizeof(bindsockaddress));
  if (status<0 && ++nport<lastport) goto TRY_NEXT_PORT;
  if (status<0) { perror("Err binding socket\n"); return -4; }

  /* Unblock socket */
  status = ioctl (sock, FIONBIO, &unblock);
  if (status<0) { perror("Err unblocking socket\n"); return -2; }

  /* set socket size to the larger of 10M and the system default */
  len=4; status = getsockopt (sock, SOL_SOCKET, SO_RCVBUF, &retbufsize, &len);
  if (status==0 && retbufsize>bufsize) bufsize=retbufsize;

  len = sizeof(unsigned int);
  status = setsockopt (sock, SOL_SOCKET, SO_RCVBUF, &bufsize, len);
  if (status<0) { perror("Err sizing receive socket buffer\n"); return -3; }

  len=4; status = getsockopt (sock, SOL_SOCKET, SO_RCVBUF, &retbufsize, &len);
  if (retbufsize<bufsize) printf ("RCV buffer is only %d bytes\n",retbufsize);

  /* Set & Verify socket has multicast interface --may not need because
     it seems this is only used for transmitting multicast data */
     
  /* status = setsockopt(sock,IPPROTO_IP,IP_MULTICAST_IF,
                      &mreq.imr_interface.s_addr,
                      sizeof(mreq.imr_interface.s_addr));
  if (status<0) { perror("Err setting multicast interface\n"); return -9;}
  
  status = getsockopt(sock,IPPROTO_IP,IP_MULTICAST_IF,&if_addr,&if_siz);
  if (status<0) { perror("No multicast interface for socket\n"); return -8; }
  printf("Interface address received = %d \n",if_addr);  */

  return sock;
}

void closesocket (int sock,char *mcaddress, char *ifaddress) {
/******** NOTE:  Closing of a socket removes all multicast membership,
 ********        so the steps below are not necessary, but just here
 ********        as a coding exercise (and an example for use elsewhere */

  struct ip_mreq mreq;               /* Multicast request structure */
  int status;

  /* Convert network addresses to binary form and place in mreq structure */
  inet_aton(mcaddress,&mreq.imr_multiaddr);
  inet_aton(ifaddress,&mreq.imr_interface);

  /* Now remove membership from multicast group. */
  status = setsockopt(sock,IPPROTO_IP,IP_DROP_MEMBERSHIP,&mreq,sizeof(mreq));

  close (sock);
}

/* Receives a packet */
int recvpacket (int sock, char *pack, int length, int debug) {
#include <errno.h>
  ssize_t status;
  int addrlen=16;  
  extern int errno;
  
  status = recvfrom(sock,pack,length,0,0,(socklen_t *)&addrlen);
  
  if (status <= 0) {
     if (errno == EAGAIN) {
       /* printf("No valid data coming in error\n"); */
     }  
     else  {  
       printf("Sock = %d, length = %d \n",sock,length);
       printf("recvfrom status  = %d\n",status);
       perror("");
     }  
  }
  else if (debug > 1) {
     printf("Received %d bytes!!\n",status);
  }
  return status;
}


/* Receives some # of sdds packets */
int recvpacketsdds (int sock, char *pack, int length, int debug) {
#include <errno.h>
  ssize_t status;
  int addrlen=16;  
  extern int errno;
  int i,sum,num,j;
  int sddsps = 1080;

  num=length/sddsps; if (num < 1) num=1;
  sum=0; i = 0;
  while (i<num) {
    status = recvfrom(sock,&pack[i*sddsps],sddsps,0,0,(socklen_t *)&addrlen);
  
    if (status <= 0) {
       if (errno == EAGAIN) {
         /* printf("No valid data coming in error\n"); */
       }  
       else  {  
         printf("Sock = %d, length = %d \n",sock,length);
         printf("recvfrom status  = %d\n",status);
         perror("");
       }  
    }
    else {
      sum += status; i++;
      if (debug > 1) printf("Received %d bytes!!\n",status);
    }
  } 
  return sum;
}
