/****************************************************************************/
/*                                                                          */
/*   LIBSERIAL: library of serial I/O routines                              */
/*              for sample usage see _BUILDTEST                             */
/*                                                                          */
/****************************************************************************/
#include "libserial.h"


/*
-----------------------------------------------------------------------
  SYNOPSIS:
    TYPE serialport_get_KEYNAME (void *ptrHandle)

  ARGUMENTS:
    void *ptrHandle - pointer to LIBSERIAL plan handle

  RETURN:
    TYPE serialport_get_KEYNAME - returns specified KEYNAME value

  PURPOSE:
    Gets an element of the SERIALPORT plan contents

-----------------------------------------------------------------------
  SYNOPSIS:
    int_4 serialport_set_KEYNAME (void *ptrHandle, TYPE val)

  ARGUMENTS:
    void *ptrHandle - pointer to LIBSERIAL plan handle
    TYPE val        - value to set key to

  RETURN:
    int_4 serialport_set_KEYNAME - returns function status (0 = noerror)

  PURPOSE:
    Sets an element of the SERIALPORT plan contents and performs updates

  REVISIONS:
    20060530  baseline
-----------------------------------------------------------------------
*/

int_4 serialport_set_carriage(void *ptrHandle, int_4 i4val)
{
  LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
  P->carriage = i4val;
  return(0);
}

int_4 serialport_get_debuglevel(void *ptrHandle)
{
  LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
  return(P->debuglevel);
}

int_4 serialport_set_debuglevel(void *ptrHandle, int_4 i4val)
{
  LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
  P->debuglevel = i4val;
  return(0);
}

int_4 serialport_get_iodelay(void *ptrHandle)
{
  LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
  return(P->iodelay);
}

int_4 serialport_set_iodelay(void *ptrHandle, int_4 i4val)
{
  LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
  P->iodelay = i4val;
  P->readdelay = i4val;
  P->writedelay = i4val;
  return(0);
}

int_4 serialport_get_readdelay(void *ptrHandle)
{
  LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
  return(P->readdelay);
}

int_4 serialport_set_readdelay(void *ptrHandle, int_4 i4val)
{
  LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
  P->readdelay = i4val;
  return(0);
}

int_4 serialport_get_writedelay(void *ptrHandle)
{
  LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
  return(P->writedelay);
}

int_4 serialport_set_writedelay(void *ptrHandle, int_4 i4val)
{
  LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
  P->writedelay = i4val;
  return(0);
}



/*
-----------------------------------------------------------------------
  SYNOPSIS:
    int_4 serialport_create(void **ptrHandle)

  ARGUMENTS:
    void **ptrHandle - pointer to serial I/O object handle

  RETURNS:  
    0 if no errors
   -1 for errors         

  PURPOSE:
    Creates an empty, initialized handle to terminal I/O object

  REVISIONS:
  2007/11/01    rdh     baseline
-----------------------------------------------------------------------
*/

int_4 serialport_create(void **ptrHandle)
{
    int_4 i;
    LIBSERIAL_Object *SP = NULL;

    /* Create handle to empty object */
    SP = (LIBSERIAL_Object *) malloc(sizeof(LIBSERIAL_Object));
    if (SP == NULL) return (-1);

    /* Default field values */
    SP->debuglevel = -1;
    SP->quiet      = -1;
    SP->fd         = -1;
    SP->readdelay  = 100000;
    
    for (i=0; i<MAX_SERIALIO_STRINGLEN; i++) SP->devname[i] = '\0';

    /* Return the handle */
    *ptrHandle = (void *) SP;

    return (0);
}


/*
-----------------------------------------------------------------------
  SYNOPSIS:
    int_4 serialport_get(void *ptrHandle)

  ARGUMENTS:
    void *ptrHandle - pointer to serial I/O object

  RETURNS:  
    0 if no errors
   -1 for errors         

  PURPOSE:
    Gets the terminal I/O (termios) settings

  REVISIONS:
  2007/11/01    rdh     baseline
-----------------------------------------------------------------------
*/

int_4 serialport_get(void *ptrHandle) 
{
    LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
    struct termios port_termios;
    int fd = (P->fd);

    /* Check for valid open file descriptor */
    if (fd <= 0) {
      if (P->debuglevel > 0)
        printf("serialport_get: invalid file/port descriptor \n");
      return (-1);
    }

    /* Get port attributes into termios structure */
    bzero(&port_termios, sizeof(port_termios));
    tcgetattr(fd, &port_termios);
    if (P->debuglevel > 1) {
      printf("cflag=%08x\n", port_termios.c_cflag);
      printf("oflag=%08x\n", port_termios.c_oflag);
      printf("iflag=%08x\n", port_termios.c_iflag);
      printf("lflag=%08x\n", port_termios.c_lflag);
      printf(" line=%02x\n", port_termios.c_line);
    }

    return (0);
}


/*
-----------------------------------------------------------------------
  SYNOPSIS:
    int_4 serialport_termios(void *ptrHandle) 

  ARGUMENTS:
    void *ptrHandle - pointer to serial I/O object

  RETURNS:  
    0 if no errors
   -1 for errors         

  PURPOSE:
    Sets the terminal I/O (termios) knobs to desired settings   

  REVISIONS:
  2007/11/01    rdh     baseline
-----------------------------------------------------------------------
*/

int_4 serialport_termios(void *ptrHandle) 
{
    LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
    int fd = (P->fd);
    struct termios port_termios;

    /* Check for valid open file descriptor */
    if (fd <= 0) {
      if (P->debuglevel > 0)
        printf("serialport_termios: invalid file/port descriptor \n");
      return (-1);
    }

    /* Get unmodified port attributes */
    bzero(&port_termios, sizeof(port_termios));
    tcgetattr(fd, &port_termios);
    if (P->debuglevel > 1) {
      printf("old cflag=%08x\n", port_termios.c_cflag);
      printf("old oflag=%08x\n", port_termios.c_oflag);
      printf("old iflag=%08x\n", port_termios.c_iflag);
      printf("old lflag=%08x\n", port_termios.c_lflag);
      printf("old  line=%02x\n", port_termios.c_line);
    }

    /* Change port settings here */
    /* port_termios.c_cflag = B9600 | CS8 | CREAD | CLOCAL | CRTSCTS; */
    port_termios.c_cflag = B9600 | CS8 | CREAD | CLOCAL; 
    port_termios.c_cc[VMIN] = 1;
    cfsetospeed(&port_termios, B9600);
    cfmakeraw(&port_termios);
    tcflush(fd, TCIFLUSH);
    tcsetattr(fd, TCSANOW, &port_termios);

    /* Output debug information */
    if (P->debuglevel > 1) {
      printf("new cflag=%08x\n", port_termios.c_cflag);
      printf("new oflag=%08x\n", port_termios.c_oflag);
      printf("new iflag=%08x\n", port_termios.c_iflag);
      printf("new lflag=%08x\n", port_termios.c_lflag);
      printf("new  line=%02x\n", port_termios.c_line);
    }

    return (0);
}


/*
-----------------------------------------------------------------------
  SYNOPSIS:
    int_4 serialport_init(void *ptrHandle, const char *DeviceName) 

  ARGUMENTS:
    void *ptrHandle  - pointer to serial I/O object
    const char *DeviceName - Linux serial device name (/dev/ttyS0, etc.)

  RETURNS:  
    0 if no errors
   -1 if port open fails

  PURPOSE:
    Opens the serial device with desired terminal parameters and 
    creates a handle to the device for serial I/O.

  REVISIONS:
  2007/11/01    rdh     baseline
-----------------------------------------------------------------------
*/

int_4 serialport_init(void *ptrHandle, const char *DeviceName) 
{
    LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
    int fd = -1;       

    /* Open the terminal port */
    P->fd = fd;
    strncpy(P->devname,DeviceName,(size_t)MAX_SERIALIO_STRINGLEN);
    fd = open(P->devname, O_RDWR | O_NOCTTY | O_NDELAY | O_NONBLOCK);
    if (fd < 0) {
      if (P->debuglevel > 0)
        printf("serialport_init: error %d %s\n", errno, strerror(errno));
      return (-1);
    }

    /* Set the terminal device ownership */
    if (fcntl(fd, F_SETOWN, getpid()) == -1) {
      if (P->debuglevel > 0)
        printf("serialport_init: error failed to set device ownership\n");
      if (fd > 0) close(fd);
      return (-1);
    }
   
    /* Assign the valid file descriptor to the handle */
    P->fd = fd;
    if (P->debuglevel > 0) {
      printf("serialport_init: DeviceName=%s\n", P->devname);
      printf("serialport_init: PortDesc=%d\n", P->fd);
    }

    /* Set up the terminal characteristics */
    if (serialport_termios(ptrHandle) < 0) {
      if (P->debuglevel > 0)
        printf("serialport_init: error setting terminal properties \n");
      if (fd > 0) close(fd);
      return (-1);
    }

    return (0);
} 


/*
-----------------------------------------------------------------------
  SYNOPSIS:
    int_4 serialport_write(void *ptrHandle, char *bufWrite)   

  ARGUMENTS:
    void *ptrHandle  - pointer to serial I/O object
    char *bufWrite   - string to write to serial device

  RETURNS:  
   >0 number of characters successfully written to device
   -1 write failed

  PURPOSE:
    Writes null terminated string to the serial port and returns 
    the status 

  REVISIONS:
  2007/11/01    rdh     baseline
-----------------------------------------------------------------------
*/

int_4 serialport_write(void *ptrHandle, char *bufWrite)
{
    LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
    int_4 numWrite;
    int fd = (P->fd);

    /* Check for valid open file descriptor */
    if (fd <= 0) {
      if (P->debuglevel > 0)
        printf("serialport_write: invalid file/port descriptor \n");
      return (-1);
    }

    /* Pause between commands for serial device to catch up */
    if ((P->writedelay) > 0) usleep(P->writedelay);

    /* Write the string to the serial device */
    numWrite = write(fd, bufWrite, strlen(bufWrite));
    if (numWrite <= 0) {
      printf("serialport_write: error %d %s\n", errno, strerror(errno));
      return (-1);
    }

    /* Debug output */
    if (P->debuglevel > 0) {
      printf("serialport_write: wrote %4.4d chars: %s\n", numWrite, bufWrite);
    }

    return (numWrite);
}


/*
-----------------------------------------------------------------------
  SYNOPSIS:
    int_4 serialport_read(void *ptrHandle, char *bufRead, int_4 bufLen)

  ARGUMENTS:
    void *ptrHandle - pointer to serial I/O object
    char  *bufRead  - string read from serial device 
    int_4 bufLen    - length of string read from serial device

  RETURNS:  
  >=0 no errors 
   -1 read failed

  PURPOSE:
    Read a string from the serial port 

  REVISIONS:
  2007/11/01    rdh     baseline
-----------------------------------------------------------------------
*/

int_4 serialport_read(void *ptrHandle, char *bufRead, int_4 bufLen)
{
    LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
    int numRead;
    int fd = (P->fd);

    /* Initialize return string in case of failure */
    bufRead[0] = '\0';

    /* Check for valid open file descriptor */
    if (fd <= 0) {
      if (P->debuglevel > 0)
        printf("serialport_read: invalid file/port descriptor \n");
      return (-1);
    }

    /* Give serial port time to respond */
    if ((P->readdelay) > 0) usleep(P->readdelay);

    /* Read string from serial port */
    numRead = read(fd, bufRead, (bufLen-1));
    if (numRead < 0) {
      /* It is possible that there was simply no data to read */
      if (errno == EAGAIN) {
        if (bufLen > 0) bufRead[0] = '\0';
        return (0); 
      }
      else {
        if (P->debuglevel > 0)
          printf("serialport_read: error %d %s\n", errno, strerror(errno));
          return (-1);
      }
    }

    /* Terminate string */
    numRead = (numRead<bufLen) ? (numRead) : (bufLen);
    bufRead[numRead] = '\0';

    /* Debug output */
    if (P->debuglevel > 0) {
      printf("serialport_read:   read %4.4d chars: %s\n", numRead, bufRead);
    }

    return (numRead);
} 


/*
-----------------------------------------------------------------------
  SYNOPSIS:
    int_4 serialport_close(void *ptrHandle)

  ARGUMENTS:
    void *ptrHandle - pointer to serial I/O object

  RETURNS:  
    0 no errors
   -1 failed to close

  PURPOSE:
    Close an open serial port          

  REVISIONS:
  2007/11/01    rdh     baseline
-----------------------------------------------------------------------
*/

int_4 serialport_close(void *ptrHandle)
{
    LIBSERIAL_Object *P = (LIBSERIAL_Object *) ptrHandle;
    int fd = (P->fd);

    /* Check for valid open file descriptor */
    if (fd <= 0) {
      if (P->debuglevel > 0)
        printf("serialport_close: invalid file/port descriptor \n");
      return (-1);
    }

    /* Close the port */
    if (close(fd) != 0) {
      if (P->debuglevel > 0)
        printf("close error %d %s\n", errno, strerror(errno));
      return (-1);
    }
    P->fd = -1;
    
    return (0);
} 


/*
-----------------------------------------------------------------------
  SYNOPSIS:
    int_4 serialport_destroy(void **ptrHandle)

  ARGUMENTS:
    void **ptrHandle - pointer to serial I/O object handle

  RETURNS:  
    0 if no errors
   -1 for errors         

  PURPOSE:
    Destroys handle to terminal I/O object

  REVISIONS:
  2007/11/01    rdh     baseline
-----------------------------------------------------------------------
*/

int_4 serialport_destroy(void **ptrHandle)
{
    LIBSERIAL_Object *SP = (LIBSERIAL_Object *) *ptrHandle;

    /* Close the serial port */
    if (SP == NULL) return (-1);
    serialport_close(SP);

    /* Destroy buffers, contents, and handle */
    if (SP != NULL) free(SP);

    return (0);
}




#ifdef _BUILDTEST
int_4 main(int argc, char *argv[])
{
    void *SP; 
    int  iSpot;
    char sCmd[MAX_SERIALIO_STRINGLEN];
    char sResult[MAX_SERIALIO_STRINGLEN];

    if (argc < 2 || argc > 2)
    {
        printf("\n Usage: serialio <devicename> \n\n");
        return (0);
    }

    /* Create, setup, and open serial port */
    serialport_create(&SP);
    serialport_set_debuglevel(SP,1);
    serialport_set_iodelay(SP,75000);
    if (serialport_init(SP,argv[1]) < 0) {
      serialport_destroy(&SP);
      return (0);
    }
    printf("Debuglevel: %i\n",serialport_get_debuglevel(SP));
    printf("ReadDelay:  %i\n",serialport_get_readdelay(SP));
    printf("WriteDelay: %i\n",serialport_get_writedelay(SP));

    printf("\n\n *** Type q to quit ***\n\n");
    while (1)
    {
            /* Get the command */
            printf("Command: ");
            gets(sCmd);
            if (sCmd[0] == 'q' || sCmd[0] == 'Q') return (0);
            iSpot = strlen(sCmd);

            /* Terminate the command string with <CR><LF><NULL> */
            sCmd[iSpot]   = 0x0d; 
            sCmd[iSpot+1] = 0x0a; 
            sCmd[iSpot+2] = 0x00;

            /* Write to serial port */
            serialport_write(SP,sCmd);

            /* Read from serial port */
            serialport_read(SP,sResult,MAX_SERIALIO_STRINGLEN);
    } 

    /* Free port resources */
    serialport_destroy(&SP);

} 
#endif
